/* 
** Module   : FillInConfig.java
** Abstract : stores FillIn specific state
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050725   @21810 Creation based on general concept.
** 002 SIY 20050831   @22502 Refactoring (see @@22394-22409). 
** 003 SIY 20050907   @22635 Fixed use of @inheritDoc tag in javadoc
** 004 SIY 20050913   @22838 Removed redundant disableAutoZap attribute,
**                           fixed handling of the cursorOffset, 
**                           moved getDataType() to superclass, cleanups.
** 005 SIY 20060209   @24467 Major refactoring.
** 006 SIY 20060324   @25281 Added isAutoReturn().
** 007 SIY 20060404   @25258 Support for the text group.
** 008 SIY 20060428   @25086 Moved some members to ControlTextConfig.
** 009 GES 20060803   @28406 Made externalizable for better performance.
**                           Many code standard cleanups.
** 010 SIY 20060926   @30023 Added getters for the next/previous ID's.
** 011 SIY 20100619          Switching to own implementation of character UI.
** 012 SIY 20100804          Changes in imports, some refactorings.
** 013 SVL 20101006          Added NATIVE attribute.
** 014 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 015 SIY 20111202          Updated imports.
** 016 CS  20121031          Removed client specific references
** 017 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 018 CA  20141206          Added support for SELECTION-START/END, needed for GUI FILL-IN.
** 019 CA  20141220          Moved some client-side state common for both ChUI and GUI from the 
**                           FillIn class.   This exists only on client-side, and is not 
**                           transferred during widget config sync.
** 020 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 021 HC  20150323          Removed business logic from config classes.
** 022 SVL 20150605          Added zapState.
*/

package com.goldencode.p2j.ui;

import java.io.*;

import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.util.*;

/**
 * Contains FillIn specific widget state.
 */
public class FillInConfig
extends ControlTextConfig
{
   /** Constant identifying the selection start or end was not set. */
   public static final int INVALID_SELECTION = Integer.MIN_VALUE;

   /** The Progress CURSOR-OFFSET attribute. */
   public int cursorOffset = 0;
   
   /** Determines if AUTO-ZAP is allowed. */
   public boolean autoZap = true;

   /** Current AUTO-ZAP state. */
   public boolean zapState = true;
   
   /** ID of previous widget in the group. */
   public int prevId = -1;

   /** ID of next widget in the group. */
   public int nextId = -1;

   /** The Progress DEBLANK attribute (not implemented). */
   public boolean deblank = false;

   /** The Progress AUTO-RETURN attribute. */
   public boolean autoReturn = false;
   
   /** Group mode flag. */
   public boolean grouping = false;

   /** The Progress NATIVE attribute. */
   public boolean nativ = false;
   
   /** Position of the selection start. */
   public int selectionStart = INVALID_SELECTION;
   
   /** Position of the selection end. */
   public int selectionEnd = INVALID_SELECTION;
   
   // START_CLIENT_SIDE_MARKER
   // on client-side, following are fields containing shared state between different widget 
   // implementations (i.e. GUI and ChUI)
   /**
    * Internal presentation of displayed data. May be <code>null</code>
    * which means that value can't be represented with the current format.
    * Such value is rendered as several question marks and can only be reset
    * to a valid value by ZAP or CLEAR actions. At resetting corresponding
    * Presentation instance is generated and is assigned to this member.
    */
   public transient DisplayFormat.Presentation appDataPres;
   
   /** Instance of display format. */ 
   public transient DisplayFormat appFormat;

   /** Alternative variable to display. */
   public transient BaseDataType displayVar = null;
   
   /** Alternative format to display. */ 
   public transient DisplayFormat displayFormat = null;
   
   /** Alternative data presentation to display. */
   public transient DisplayFormat.Presentation displayDataPres = null;
   
   /** Widget in focus flag */
   public transient boolean widgetFocused = false;
   // END_CLIENT_SIDE_MARKER

   /**
    * Default constructor (only used in de-serialization).
    */
   public FillInConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected FillInConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   public FillInConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      FillInConfig cfg = (FillInConfig) config;
      
      cursorOffset = cfg.cursorOffset;
      autoZap = cfg.autoZap;
      zapState = cfg.zapState;
      prevId = cfg.prevId;
      nextId = cfg.nextId;
      deblank = cfg.deblank;
      autoReturn = cfg.autoReturn;
      grouping = cfg.grouping;
      nativ = cfg.nativ;
      selectionStart = cfg.selectionStart;
      selectionEnd = cfg.selectionEnd;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      cursorOffset   = in.readInt();
      prevId         = in.readInt();
      nextId         = in.readInt();
      autoZap        = in.readBoolean();
      zapState       = in.readBoolean();
      deblank        = in.readBoolean();
      autoReturn     = in.readBoolean();
      grouping       = in.readBoolean();
      nativ          = in.readBoolean();
      selectionStart = in.readInt();
      selectionEnd   = in.readInt();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeInt(cursorOffset);
      out.writeInt(prevId);
      out.writeInt(nextId);
      out.writeBoolean(autoZap);
      out.writeBoolean(zapState);
      out.writeBoolean(deblank);
      out.writeBoolean(autoReturn);
      out.writeBoolean(grouping);
      out.writeBoolean(nativ);
      out.writeInt(selectionStart);
      out.writeInt(selectionEnd);
  }
}
