/*
** Module   : FillInWidget.java
** Abstract : server side FillIn widget implementation
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20051026   @23490 Created initial version.
** 002 GES 20060114   @23909 Removed totally useless screen-value attr.
** 003 SIY 20060404   @25357 Support for the text group.
** 004 GES 20060803   @28375 Cleanup and refactoring.
** 005 GES 20080320   @37590 Added TYPE attribute.
** 006 SVL 20101006          Added NATIVE attribute.
** 007 SIY 20111101          Updated imports.
** 008 CA  20130221          Refactored usage of TYPE attribute.
** 009 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 010 CA  20131115          Changes related to properly instantiate dynamic resources.
** 011 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 012 SVL 20141102          Make this class implement AutoZapElement.
** 013 GES 20141126          Added SET-SELECTION() method.
** 014 SVL 20150605          Properly implemented AUTO-ZAP and CURSOR-OFFSET.
*/ 

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Server side FillIn widget implementation.
 */
@LegacyResource(resource = LegacyResource.FILL_IN)
public class FillInWidget
extends ControlTextWidget<FillInConfig>
implements AutoZapElement
{
   /**
    * Default constructor.
    */
   public FillInWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public FillInWidget(boolean dynamic)
   {
      super(dynamic, new FillInConfig());
   }

   /**
    * Get value of the AUTO-ZAP attribute of the focused fill-in.
    *
    * @return   Value of the AUTO-ZAP attribute of the focused fill-in. If there is no focused
    *           fill-in, <code>true</code> is returned.
    */
   static logical isFocusedAutoZap()
   {
      FillInWidget focused = getFocusedFillIn();
      return focused != null ? new logical(focused.config.zapState) : new logical(true);
   }

   /**
    * Get focused fill-in.
    *
    * @return focused fill-in widget or <code>null</code> if there is no focused fill-in.
    */
   private static FillInWidget getFocusedFillIn()
   {
      handle h = LogicalTerminal.focus();
      if (!h.isUnknown())
      {
         WrappedResource obj = h.getResource();
         if (obj.getClass().isAssignableFrom(FillInWidget.class))
            return (FillInWidget) obj;
      }

      return null;
   }

   /**
    * Sets the AUTO-RETURN option.
    * 
    * @param    autoReturn
    *           <code>true</code> if the option is set to ON.
    */
   public void setAutoReturn(boolean autoReturn)
   {
      config.autoReturn = autoReturn;
      pushScreenDefinition();
   }

   /**
    * Sets the DEBLANK option.
    * 
    * @param    deblank
    *           <code>true</code> if the option is set to ON.
    */
   public void setDeblank(boolean deblank)
   {
      // defined only for fill-ins
      config.deblank = deblank;
      pushScreenDefinition();
   }

   /**
    * Sets a read only state.
    * 
    * @param   newReadOnly
    *          Is widget read only.
    */
   public void setReadOnly(boolean newReadOnly)
   {
      config.readOnly = newReadOnly;
   }

   /**
    * Start text group of widgets (word wrap mode).
    * 
    * @param   prev
    *          Reference to previous fill-in widget or null if no such
    *          widgets.
    */
   public void startGrouping(FillInWidget prev)
   {
      config.grouping = true;
      
      if (prev != null)
      {
         prev.setNextGroupItem(this);
         config.prevId = prev.getId();
      }
   }

   /**
    * Link widget with the next widget in the text group.
    * 
    * @param   next
    *          Reference to next widget in group.
    */
   public void setNextGroupItem(FillInWidget next)
   {
      if (next != null)
         config.nextId = next.getId();
   }

   /**
    * Reset grouping flag and links to previous and next widget in the text
    * group.
    */
   public void resetGrouping()
   {
      config.grouping = false;
      config.nextId = -1;
      config.prevId = -1;
   }

   /**
    * Set value of the NATIVE attribute.
    *
    * @param   nativ
    *          New value for the attribute.
    */
   public void setNative(boolean nativ)
   {
      config.nativ = nativ;
   }

   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(int64 start, int64 end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(int64 start, int end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(int start, int64 end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(int start, int end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Sets the DISABLE-AUTO-ZAP option.
    * 
    * @param    disableAutoZap
    *           <code>true</code> if the option is set to ON.
    */
   public void setDisableAutoZap(boolean disableAutoZap)
   {
      config.autoZap = !disableAutoZap;
      pushScreenDefinition();
   }

   /**
    * Set AUTO-ZAP attribute.
    * 
    * @param    value
    *           New value for the attribute.
    */
   public void setAutoZap(boolean value)
   {
      setAutoZap(new logical(value));
   }

   /**
    * Set AUTO-ZAP attribute.
    *
    * @param    autoZap
    *           New value for the attribute.
    */
   public void setAutoZap(logical autoZap)
   {
      setAutoZap(autoZap, name().toStringMessage());
   }

   /**
    * Set value of the AUTO-ZAP attribute.
    *
    * @param    autoZap
    *           Value of the attribute.
    */
   public void setAutoZap(final logical autoZap, String widgetName)
   {
      if (autoZap.isUnknown())
      {
         ErrorManager.recordOrShowError(4083,
                  "**Unable to assign UNKNOWN value to attribute AUTO-ZAP on FILL-IN " +
                        widgetName, false, false);
         return;
      }

      ConfigManager.syncConfigChanges(new Runnable()
      {
         @Override
         public void run()
         {
            config.zapState = autoZap.booleanValue();
         }
      });

      pushFillInDefinition();
   }

   /**
    * Get value of the AUTO-ZAP attribute.
    *
    * @return   value of the attribute.
    */
   public logical isAutoZap()
   {
      // Oddly enough, 4GL returns zap state of the FOCUSED fill-in, not necessarily this one
      return isFocusedAutoZap();
   }

   /**
    * Get value of the CURSOR-OFFSET attribute.
    *
    * @return  value of the attribute.
    */
   public integer getCursorOffset()
   {
      if (isFocused())
      {
         // cursor can take position after the last character of the fill-in
         int offset = Math.min(config.cursorOffset + 1, (int) config.widthChars);
         return new integer(offset);
      }
      else
      {
         return new integer();
      }
   }

   /**
    * Sets the position of the cursor. Corresponds CURSOR-OFFSET attribute.
    *
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   public void setCursorOffset(double offset)
   {
      setCursorOffset(new decimal(offset));
   }

   /**
    * Sets the position of the cursor. Corresponds CURSOR-OFFSET attribute.
    *
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   public void setCursorOffset(NumberType offset)
   {
      setCursorOffset(offset, name().toStringMessage());
   }

   /**
    * Sets the position of the cursor. Corresponds CURSOR-OFFSET attribute.
    *
    * @param    offset
    *           The character index at which the cursor is to be placed.
    * @param    widgetName
    *           Widget name. Used for error handling.
    */
   public void setCursorOffset(NumberType offset, String widgetName)
   {
      if (!isFocused())
         return;

      if (offset.isUnknown())
      {
         ErrorManager.recordOrShowError(4083,
               "**Unable to assign UNKNOWN value to attribute CURSOR-OFFSET on FILL-IN " +
                  widgetName, false, false);
         return;
      }

      final int val = offset.intValue();
      if (val < 1 || val > config.widthChars)
      {
         ErrorManager.recordOrShowError(4057,
               String.format("**Attribute CURSOR-OFFSET for the FILL-IN %s has an invalid " +
                     "value of %d", widgetName, val),
               false, false);
         return;
      }

      ConfigManager.syncConfigChanges(new Runnable()
      {
         @Override
         public void run()
         {
            config.cursorOffset = val - 1;
         }
      });

      pushFillInDefinition();
   }

   /**
    * Push screen definition instance down to the client. Supports both fill-ins for in-browse
    * editing and normal fill-ins.
    *
    */
   private void pushFillInDefinition()
   {
      if (frame != null)
      {
         pushScreenDefinition();
      }
      else
      {
         ScreenDefinition def = new ScreenDefinition();
         def.addConfig(config);
         LogicalTerminal.pushScreenDefinition(new ScreenDefinition[]{def});
      }
   }

   /**
    * Check if this fill-in is focused.
    *
    * @return <code>true</code> if this fill-in is focused.
    */
   private boolean isFocused()
   {
      handle h = LogicalTerminal.focus();
      if (!h.isUnknown())
      {
         WrappedResource obj = h.getResource();
         if (obj.equals(this))
            return true;
      }

      return false;
   }
}
