/*
** Module   : FocusManager.java
** Abstract :
**
** Copyright (c) 2010-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100915 Created initial version
** 002 SIY 20101015 Changes related to refactoring of Screen.
** 003 SIY 20101122 Switching from KeyEvent to KeyInput.
** 004 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 005 CA  20140805 Explicitly import any chui package classes.
** 006 VIG 20140919 Changed status method call.
** 007 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 008 HC  20150221 Improvements to frame tab order.
** 009 VIG 20150224 Added menu support to getFirstEnabledWidget
** 010 SVL 20150608 Reflect API change in fill-in.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Focus manager.
 * 
 * @author 	SIY
 * @version 1.0
 */
public class FocusManager
{
   /** State of the focus on entry to waitFor: undefined */
   public static final int FOCUS_UNDEF = 0;

   /** State of the focus on entry to waitFor: invalid */
   public static final int FOCUS_INVALID = 1;
   
   /** State of the focus on entry to waitFor: valid */
   public static final int FOCUS_VALID = 2;
   
   /** 
    * State of the focus on entry to waitFor: hidden, as in not visible.
    * This is not the HIDDEN flag state.
    */
   public static final int FOCUS_HIDDEN = 3;

   /** Owner {@link ThinClient} instance. */
   private final ThinClient tc;

   /** Do not adjust focus in WAIT-FOR if set */
   private boolean noFocusAdjustment = false;

   /** Lookup mode for focusable widgets in frames: wrap */
   public static final boolean FOCUS_WRAP   = true;

   /** Lookup mode for focusable widgets in frames: no wrap */
   public static final boolean FOCUS_NOWRAP = false;

   /**
    * Constructor.
    * 
    * @param   tc
    *          Reference to {@link ThinClient}.
    */
   public FocusManager(ThinClient tc) 
   {
      this.tc = tc;
   }
   
   /**
    * Adjusts focus if it is either undefined or invalid and a trigger has
    * enabled at least one widget.
    */
   public void adjustFocus()
   {
      // get the current focus
      Widget focus = UiUtils.getCurrentFocus();
      Widget next  = null;
   
      // get the focus state: either undefined or invalid
      int focusState = focusState(focus);
   
      // if focus is invalid, the next enabled widget in the frame gets it
      if (focusState == FOCUS_INVALID || focusState == FOCUS_HIDDEN)
      {
         Frame frame = (Frame)UiUtils.locateFrame(focus);
         // if invalid, search for next enabled widget, with wrapping
         next = frame.getNextEnabledWidget(focus, FocusManager.FOCUS_WRAP);
   
         if (next != null)
         {
            // set focus on the next enabled widget silently
            setFocusOn(next);
   
            return;
         }
         
         // if focus is disabled and there is no other enabled widget, 
         // then focus is not adjusted if other widget type beside combo or 
         // fill-in is in focus and the last trigger returned normal
         boolean b = tc.isTriggerWasAppliedWithoutWaitFor();
         if (b && !(focus instanceof FillIn || focus instanceof ComboBox))
         {
            // this is only if there was no nested wait-for loop executed
            // by the trigger
            
            if (frame.hasEnabled())
               return;
         }
      }
   
      // if focus is undefined, the very first enabled widget gets it
      next = getFirstEnabledWidget();
   
      if (next == null)
      {
         // still nothing
         // set focus to undefined, if no enabled widget found
         Window.resolveWindow().setFocus(null);
         // set the status to default if there is no enabled widget
         Window.resolveWindow().status(null);
         return;
      }
   
      // set focus on the first enabled widget silently
      setFocusOn(next);
   
      // adjust zap and insert state as if ENTRY was sent
      if (focus instanceof FillIn && 
          next instanceof FillIn  && 
          focusState == FOCUS_HIDDEN)
      {
         if (((FillIn) focus).isZapActive())
         {
            FillIn fNext  = (FillIn) next;
            
            fNext.setZapMode(true);
            fNext.setInsertMode(false);
         }
      }
   }

   /**
    * Sets focus on the specified widget silently and unconditionally.
    *
    * @param  widget
    *         widget to set focus on; it should be a valid widget, but
    *         can be disabled  
    */
   public void setFocusOn(final Widget widget)
   {
      Widget oldFocus = UiUtils.getCurrentFocus();
      boolean ins = Window.getInsertMode() == Window.INSERT_MODE_INS;
      tc.eventDrawingBracket(widget, false, false, new Runnable()
      {
         @Override
         public void run()
         {
            FocusManager.setInitialFocus(widget);
         }
      });
   
      Widget newFocus = UiUtils.getCurrentFocus();
      if (newFocus != null &&
          newFocus instanceof FillIn)
      {
         FillIn newFillIn = (FillIn) newFocus;
   
         if (oldFocus == null)
         {
            newFillIn.setInsertMode(ins);
         }
         else if (newFocus != oldFocus &&
                  (!oldFocus.isEnabled() || !oldFocus.isVisible()))
         {
            if (oldFocus instanceof Editor)
            {
               newFillIn.setInsertMode(false);
            }
            else
            {
               newFillIn.setInsertMode(ins);
   
               if (oldFocus instanceof FillIn &&
                   !((FillIn) oldFocus).isZapActive() &&
                   !newFillIn.isLinkedToBrowse())
               {
                  newFillIn.setPendingNoZap();
               }
            }
         }
      }
   }

   /**
    * Looks up the very first enabled widget among all displayed frames and menubar.
    *
    * @return the first enabled widget or <code>null</code>
    */
   public Widget getFirstEnabledWidget()
   {
      // loop through the displayed frames
      Frame frame = UiUtils.getFirstVisibleFrame();
   
      while (frame != null)
      {
         if (frame.isAlive())
         {
            Widget comp = frame.getFirstEnabledWidget();
            
            if (comp != null) {
               return comp;
            }
         }
   
         // try next frame
         frame = UiUtils.getNextVisibleFrame(frame);
      }
   
      // no enabled widgets in all frames
      
      if (frame == null)
      {
         Window<?> wnd = Window.resolveWindow();
         if (wnd.config.menubarId != 0)
         {
            WidgetRegistry<?> registry = tc.getOutputManager().getRegistry();
            Menu<?> menu = (Menu<?>) registry.getComponent(wnd.config.menubarId);
            return menu.getFirstEnabledWidget();
         }
      }
      
      return null; 
   }

   /**
    * Sets initial focus for the WAIT-FOR with invalid focus on entry
    * and no explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @return <code>true</code> if WAIT-FOR can proceed
    */
   public boolean focusInvalidImplicit(Widget inFocus)
   {
      this.noFocusAdjustment = false;
   
      // figure out the frame
      Frame frame = (Frame)UiUtils.locateFrame(inFocus);
   
      // find the first enabled widget that follows
      Widget first = frame.getNextEnabledWidget(inFocus, FocusManager.FOCUS_WRAP);
   
      if (first == null)
      {
         // no enabled widgets within the same frame; check others
         first = getFirstEnabledWidget();
   
         if (first == null)
         {
            // no enabled widgets at all; focus becomes undefined
            Window.getDefaultWindow().setFocus(null);
   
            return true;
         }
      }
   
      // set focus on the found widget
      setFocusOn(first);
   
      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with valid focus on entry
    * and an explicit FOCUS specification.
    *
    * @param    inFocus
    *           widget in focus
    * @param    toFocus
    *           candidate widget for receiving focus
    * @param    list
    *           list of event to wait for 
    * @param    cop
    *           currently executed operation
    * @param    hidden
    *           <code>true</code> if this is the case of hidden focus
    *         
    * @return   <code>true</code> if WAIT-FOR can proceed
    */
   public boolean focusValidExplicit(Widget inFocus, 
                                      Widget toFocus,
                                      EventList list, 
                                      int       cop,
                                      boolean   hidden)
   {
      this.noFocusAdjustment = false;
   
      // check the target's enabled status
      if (!toFocus.focusTraversable())
      {
         // leave the focus where it is
         setFocusOn(inFocus);
   
         return true;
      }
      
      Widget inFrame = UiUtils.locateFrame(inFocus);
      Frame toFrame = (Frame) UiUtils.locateFrame(toFocus);
      boolean sameFrame = inFrame == toFrame;
      boolean fastExit = false;
   
      boolean isUpdate = cop == EventList.UPDATE || 
                         cop == EventList.SET    || 
                         cop == EventList.PROMPT_FOR;
      boolean isWaitFor = cop == EventList.WAIT_FOR;
   
      if (isWaitFor && hidden)
      {
         Widget first = getFirstEnabledWidget();
         Frame firstFrame = (Frame) UiUtils.locateFrame(first);
   
         // if previous focus is hidden and explicit focus is the first 
         // focusable widget on screen, adjust focus to it
         if (toFocus == first)
         {
            setFocusOn(toFocus);
         }
         else
         {
            if (inFrame.isVisible() == false && inFrame == firstFrame)
            {
               // if the previous focus frame is hidden, focus is set to
               // undefined
                Window.getDefaultWindow().setFocus(null);
            }
            // else, search for focus. if explicit focus was reached, adjust
            // focus to it, but do not send ENTRY - it will be sent later
            fastExit = searchForFocusWithEntry(first, toFocus);
         }
      }
      
      // remove highlight from the previous focus
      tc.highlight(inFocus, false);
   
      Widget next = toFocus;
   
      // scan enabled widgets of the frame and chose the first accepting ENTRY
      while (next != null)
      {
         this.noFocusAdjustment = true;
         boolean state = next.focusTraversable();
   
         // send ENTRY event and see the response
         boolean check = tc.sendEntry(next);
   
         if (check || (sameFrame && next == inFocus) || fastExit)
         {
            // no-apply must be ignored if X and Y are in the same frame and
            // X is reached
            // no-apply must be ignored if it was focused due to hidden 
            // initial focus
   
            if (widgetLostFocusability(inFocus, next, state, FOCUS_VALID))
            {
               return true;
            }
   
            // ENTRY accepted
            this.noFocusAdjustment = false;
            break;
         }
         
         // ENTRY not accepted, remove highlight
         next.setHighlighted(false);
   
         // look up the next candidate
         next = toFrame.getNextEnabledWidget(next, FocusManager.FOCUS_NOWRAP);
      }
   
      if (next == null && sameFrame == false) 
      {
         // if focus was not found and X and Y are not in the same frame, then:
         // 1. if event list waits for Y frame events, end, sending leave to X
         // 2. else, focus another widget without ENTRY, starting from X
         if (isUpdate                                     ||
             list.anyEventForWidget(UiUtils.lookupWidgetId(toFocus)) ||
             list.anyEventForWidget(UiUtils.lookupWidgetId(toFrame)))
         {
            tc.sendLeave(inFocus);
   
            return false;
         }
   
         if (!searchForFocusWithoutEntry(inFocus))
         {
            // focus remains null if no widget found
            Window.getDefaultWindow().setFocus(null);
         }
   
         return true;
      }
   
      // check the success
      if (next == null)
      {
         // no widget wants focus; check event list for global events
         if (!UiUtils.checkEventList(list))
         {
            // no global events; should exit after sending LEAVE to inFocus
            tc.sendLeave(inFocus);
   
            return false;
         }
         
         // leave the focus where it is
         setFocusOn(inFocus);
   
         return true;
      }
   
      // a good candidate is found; set focus on it
      setFocusOn(next);
   
      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with undefined or hidden focus on 
    * entry and no explicit FOCUS specification.
    *
    * @param    list
    *           list of event to wait for 
    * @param    hidden
    *           <code>true</code> if this is the case of hidden initial focus
    * 
    * @return   <code>true</code> if WAIT-FOR can proceed
    */
   public boolean focusUndefImplicit(EventList list, boolean hidden)
   {
      this.noFocusAdjustment = false;
   
      // find the very first enabled widget
      Widget first = getFirstEnabledWidget();
      if (first == null)
      {
         // no enabled widgets; check event list for global events
         if (!UiUtils.checkEventList(list))
         {
            // no global events; should exit immediately
            return false;
         }
      }
   
      // found an enabled widget; figure out the owning frame
      Frame frame = (Frame)UiUtils.locateFrame(first);
   
      // find the first enabled widget of the frame which accepts ENTRY
      while (first != null)
      {
         this.noFocusAdjustment = true;
         boolean state = first.focusTraversable();
   
         // send ENTRY event and see the response
         boolean check = tc.sendEntry(first);
         
         if (check)
         {
            if (state && first.focusTraversable() != state)
            {
               if (hidden == false || 
                  searchForFocusWithoutEntry(first) == false)
               {
                  // focus was disabled during ENTRY, focus remains undefined
                  Window.getDefaultWindow().setFocus(null);
               }
   
               return true;
            }
   
            // ENTRY accepted
            this.noFocusAdjustment = false;            
            break;
         }
         
         // ENTRY not accepted, remove highlight
         first.setHighlighted(false);
   
         first = frame.getNextEnabledWidget(first, FocusManager.FOCUS_NOWRAP);
      }
   
      // end of search; either found or hit end of frame
      if (first == null)
      {
         // no enabled widgets; check event list for global events
         if (!UiUtils.checkEventList(list))
         {
            // no global events; should exit immediately
            return false;
         }
         
         // focus remains undefined
         return true;
      }
   
      // set focus on the found widget
      setFocusOn(first);
   
      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with undefined focus on entry
    * and an explicit FOCUS specification.
    *
    * @param  toFocus
    *         candidate widget for receiving focus
    * @param  list
    *         list of event to wait for
    * @param  cop
    *         currently executed operation
    * @param  explicit
    *         when <code>true</code>, it marks a real explicit focus; else, it
    *         marks explicit focus generated due to previous pending focus, in
    *         non-nested trigger mode.
    * 
    * @return <code>true</code> if WAIT-FOR can proceed
    */
   @SuppressWarnings("null")
   public
   boolean focusUndefExplicit(Widget toFocus, 
                                      EventList list, 
                                      int cop,
                                      boolean explicit)
   {
      this.noFocusAdjustment = false;
      // if focus candidate is not the very first enabled widget on the screen
      // ENTRY may be sent to the latter first
      Widget first  = getFirstEnabledWidget();
      Frame firstFrame = (Frame) UiUtils.locateFrame(first);
      Frame toFrame = (Frame) UiUtils.locateFrame(toFocus);
   
      boolean isUpdate = cop == EventList.UPDATE || 
                         cop == EventList.SET    || 
                         cop == EventList.PROMPT_FOR;
      boolean isWaitFor = cop == EventList.WAIT_FOR;
   
      // check the target's enabled status
      if (!toFocus.focusTraversable())
      {
         // find the very first enabled widget
         if (first == null)
         {
            // no enabled widgets; focus remains undefined
            return true;
         }
   
         if (isWaitFor)
         {
            
            // search for focus starting from the first widget sending ENTRY
            searchForFocusWithEntry(first, null);
            Widget focused = UiUtils.getCurrentFocus();
            
            if (focused == first && !first.focusTraversable())
            {
               // if first enabled widget on screen accepted ENTRY and got  
               // disabled, focus is undefined
               Window.getDefaultWindow().setFocus(null);
            }
            // if no focus was found, then current operation must end
            if (focused == null)  
            {
               return false;
            }
         }
         else
         {
            // set focus on the found widget without sending ENTRY, if not
            // WAIT-FOR
            setFocusOn(first);
         }
   
         return true;
      }
      
   
      // send entry to the first enabled widget on screen W if this is the
      // wait-for operation and the target focus is not W
      boolean entryFirst = cop == EventList.WAIT_FOR && 
                           first != null             && 
                           first != toFocus;
      boolean fastExit = false;
   
      if (isUpdate && tc.getTriggerNesting() == 0 && tc.getPendingFocus() != null)
      {
         Frame pendingFrame = (Frame) UiUtils.locateFrame(tc.getPendingFocus());
   
         // before sending ENTRY to any widget, adjust focus to last widget
         // which had ENTRY only if the frame is not out of scope
         if (pendingFrame != null && pendingFrame.isAlive())
         {
            if (tc.getPendingFocus().focusTraversable())
            {
               setFocusOn(tc.getPendingFocus());
            }
            // in this case, if explicit focus is disabled during ENTRY, focus
            // search starts from this 'implicit' focus
            first = tc.getPendingFocus();
         }
         
         // the pending focus is consumed on first usage
         tc.setPendingFocus(null);
      }
      
      if (entryFirst)
      {
         fastExit = searchForFocusWithEntry(first, toFocus);
      }
   
      // scan enabled widgets of the frame and chose the first accepting ENTRY
      Widget next = toFocus;
   
      if (next != null && entryFirst)
      {
         // at least one more ENTRY event will be sent, remove highlight from
         // the widget
         first.setHighlighted(false);
      }
      
      while (next != null)
      {
         this.noFocusAdjustment = true;
         boolean state = next.focusTraversable();
         
         // send ENTRY event and see the response
         boolean check = tc.sendEntry(next);
   
         if (check || fastExit)
         {
            // ignore NO-APPLY if there is a fast exit
   
            if (state && next.focusTraversable() != state)
            {
               boolean undefined = false;
               Widget nextFrame = UiUtils.locateFrame(next);
   
               if (isUpdate || (isWaitFor && firstFrame == nextFrame))
               {
                  undefined = searchForFocusWithoutEntry(first) == false;
               }
               
               if (isWaitFor && firstFrame != nextFrame)
               {
                  undefined = (first != null && 
                               first.focusTraversable()) || 
                               searchForFocusWithoutEntry(first) == false;
               }
               
               if (undefined)
               {
                  Window.getDefaultWindow().setFocus(null);
               }
               
               return true;
            }
   
            // ENTRY accepted
            this.noFocusAdjustment = false;
            break;
         }
   
         // ENTRY not accepted, remove highlight
         next.setHighlighted(false);
   
         // look up the next candidate
         next = toFrame.getNextEnabledWidget(next, FocusManager.FOCUS_NOWRAP);
      }
   
      // check the success
      if (next == null)
      {
         // if event list has events for toFocus or toFocus frame, 
         // wait-for ends, only if explicit is not caused by pending focus
         if (explicit &&
             (isUpdate                                     ||
              list.anyEventForWidget(UiUtils.lookupWidgetId(toFocus)) ||
              list.anyEventForWidget(UiUtils.lookupWidgetId(toFrame))))
         {
            if (tc.getPendingFocus() != null)
            {
               // send LEAVE, ignoring NO-APPLY
               tc.sendLeave(tc.getPendingFocus());
   
               // set focus to undefined
               Window.getDefaultWindow().setFocus(null);
               // 'consume' the focus
               tc.setPendingFocus(null);
            }
            return false;
         }
         
         // no widget wants focus; set focus on first enabled widget on screen
         // or undefined, if none found
         next = getFirstEnabledWidget();
         if (next == null)
         {
            Window.getDefaultWindow().setFocus(null);
         }
         else
         {
            setFocusOn(next);
            // although ENTRY was not accepted, focus was changed
            this.noFocusAdjustment = false;
         }
   
         return true;
      }
   
      // a good candidate is found; set focus on it
      setFocusOn(next);
   
      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with invalid focus on entry
    * and an explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @param  toFocus
    *         candidate widget for receiving focus
    * @param  list
    *         list of event to wait for
    *
    * @return <code>true</code> if WAIT-FOR can proceed
    */
   public boolean focusInvalidExplicit(Widget inFocus, Widget toFocus,
                                        EventList list)
   {
      this.noFocusAdjustment = false;
   
      Frame frame = null;
      Widget next = null;
   
      // check the target's enabled status
      if (!toFocus.focusTraversable())
      {
         if (searchForFocusWithoutEntry(inFocus) == false)
         {
            // no enabled widgets at all; focus becomes undefined
            Window.getDefaultWindow().setFocus(null);
         }
   
         return true;
      }
   
      // scan enabled widgets of the frame and chose the first accepting ENTRY
      frame = (Frame)UiUtils.locateFrame(toFocus);
      next = toFocus;
   
      while (next != null)
      {
         this.noFocusAdjustment = true;
         boolean state = next.focusTraversable();
   
         // send ENTRY event and see the response
         boolean check = tc.sendEntry(next);
   
         if (check)
         {
            // check if the widget was disabled during trigger
            if (widgetLostFocusability(inFocus, next, state, FOCUS_INVALID))
            {
               return true;
            }
   
            // ENTRY accepted
            this.noFocusAdjustment = false;
            break;
         }
         
         // ENTRY not accepted, remove highlight
         next.setHighlighted(false);
   
         // look up the next candidate
         next = frame.getNextEnabledWidget(next, FocusManager.FOCUS_NOWRAP);
      }
   
      // check the success
      if (next == null)
      {
         // no widget wants focus; check event list for global events
         if (!UiUtils.checkEventList(list))
         {
            // no global events; should exit immediately
            return false;
         }
   
         // set focus on toFocus ignoring NO-APPLY 
         setFocusOn(toFocus);
   
         return true;
      }
   
      // a good candidate is found; set focus on it
      setFocusOn(next);
   
      return true;
   }

   /**
    * If the toFocus widget lost its focusability, then:
    * <ul>
    *    <li>if the inFocus and toFocus widgets are both in the same frame,
    *        search for focus starting from:
    *          <ul>
    *             <li>toFocus, if inFocus is invalid</li>
    *             <li>inFocus, if inFocus is valid</li>
    *          </ul>
    *    </li>
    *    <li>else, focus is undefined if:
    *       <ul>
    *          <li>inFocus is invalid</li>
    *          <li>inFocus is valid and still focusable</li>
    *          <li>inFocus is valid, not focusable: focus first enabled and
    *              visible widget on screen, without ENTRY. if none found,
    *              focus is undefined.
    *          </li>
    *       </ul>
    *    </li>
    * </ul>
    * 
    * @param    inFocus
    *           the focus before the ENTRY event was triggered
    * @param    toFocus
    *           the widget which had ENTRY event
    * @param    state
    *           previous state of this widget (focusable or not)
    * @param    focusState
    *           the state of the focused widget, as before the trigger was
    *           invoked (valid or invalid).
    * 
    * @return   <code>true</code> if this widget lost focusability
    */
   boolean widgetLostFocusability(Widget inFocus,
                                          Widget toFocus,
                                          boolean   state, 
                                          int       focusState)
   {
      if (state && toFocus.focusTraversable() != state)
      {
         boolean undefined = false;
         
         if (UiUtils.locateFrame(inFocus) == UiUtils.locateFrame(toFocus))
         {
            // Y was disabled, search for focus without ENTRY
            Widget start = focusState == FOCUS_VALID ? inFocus : toFocus;
            undefined = searchForFocusWithoutEntry(start) == false;
         }
         else
         {
            // if X is still enabled and visible or no other widget to
            // set focus to, focus is undefined
            undefined = focusState == FOCUS_VALID && 
                         (inFocus.focusTraversable() ||
                         searchForFocusWithoutEntry(inFocus) == false);
         }
         
         if (undefined)
         {
            // focus remains undefined
            Window.getDefaultWindow().setFocus(null);
         }
   
         return true;
      }
      
      return false;
   }

   /**
    * Search for focus with ENTRY, starting from the specified widget,
    * without wrapping. If no enabled and visible widget which accepts entry
    * found in current frame, focus is undefined.
    * 
    * @param    start
    *           the widget from which focus search start
    * @param    toFocus 
    *           If search reaches this widget, sending ENTRY will be skipped.
    *           
    * @return   <code>true</code> if a widget was found
    */
   boolean searchForFocusWithEntry(Widget start, Widget toFocus)
   {
      Frame startFrame = (Frame) UiUtils.locateFrame(start);
      
      // send ENTRY to the specified widget
      boolean check = tc.sendEntry(start);
   
      if (check == false)
      {
         // the specified widget did not accept entry
         // send entry to next widgets, in tab-order
         Widget next = startFrame.getNextEnabledWidget(start, FocusManager.FOCUS_NOWRAP);
         while (next != null)
         {
            if (next == toFocus)
            {
               // reached explicit focus, it must receive focus before sending
               // ENTRY to it
               setFocusOn(next);
   
               return true;
            }
   
            // send ENTRY event
            check = tc.sendEntry(next);
            if (check)
            {
               // focus found
               break;
            }
            
            next = startFrame.getNextEnabledWidget(next, FocusManager.FOCUS_NOWRAP);
         }
         
         if (next == null)
         {
            // no widget accepts ENTRY in firstFrame, focus is undefined
            Window.getDefaultWindow().setFocus(null);
         }
      }
      
      return false;
   }

   /**
    * Sets initial focus for the WAIT-FOR with valid focus on entry
    * and no explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @return <code>true</code> always to indicate success
    */
   public boolean focusValidImplicit(Widget inFocus)
   {
      this.noFocusAdjustment = false;
      boolean state = inFocus.focusTraversable();
   
      // currently focused widget receives ENTRY event
      // NO-APPLY, if returned by a trigger, is ignored
      tc.sendEntry(inFocus);
   
      if (state && inFocus.focusTraversable() != state)
      {
         // if it was disabled or hidden by the trigger, readjust focus,
         // starting from current frame
         if (searchForFocusWithoutEntry(inFocus) == false)
         {
            // focus remains undefined
            Window.getDefaultWindow().setFocus(null);
         }
      }
   
      // currently focused widget remains in focus or focus is undefined
      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with valid focus on entry
    * and no explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @return <code>true</code> always to indicate success
    */
   public static int focusState(Widget inFocus)
   {
      if (inFocus == null)
      {
         return FOCUS_UNDEF;
      }
      
      if (inFocus.isVisible() == false)
      {
         return FOCUS_HIDDEN;
      }
   
      if (inFocus.isEnabled())
      {
         return FOCUS_VALID;
      }
   
      return FOCUS_INVALID;
   }

   /**
    * Search for focus without ENTRY, starting from the specified widget,
    * with wrapping in its frame. If no enabled and visible widget found in
    * current frame, search for focus starting with first frame on screen.
    * 
    * @param    start
    *           the widget from which focus search start
    * 
    * @return   <code>true</code> if a widget was found
    */
   public boolean searchForFocusWithoutEntry(Widget start)
   {
      if (start == null)
      {
         return false;
      }
      
      Frame frame = (Frame) UiUtils.locateFrame(start);
      
      if (frame.isVisible())
      {
         Widget next = start;
         
         do
         {
            if (next.focusTraversable())
            {
               // widget was found, request focus
               setFocusOn(next);
               return true;
            }
            
            next = frame.getNextEnabledWidget(next, FocusManager.FOCUS_WRAP);
         }
         while (next != null);
      }
      
      Widget first = getFirstEnabledWidget();
      if (first != null)
      {
         setFocusOn(first);
         return true;
      }
   
      // no widget was found
      return false;
   }

   /**
    * Initiates the focus change procedure by creating and handling a LEAVE
    * event and, if that event is not consumed by a trigger, proceeds with
    * the focus change.
    * 
    * @param   widget
    *          <code>Component</code> that is the candidate to receive focus.
    *          
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean tryFocusChange(Widget widget)
   {
      return tryFocusChange(widget, true, true);
   }

   /**
    * Initiates the focus change procedure by creating and handling a LEAVE
    * event and, if that event is not consumed by a trigger, proceeds with
    * the focus change.
    * <p>
    * Conditionally issues ENTRY event that follows the focus change.
    * 
    * @param   widget
    *          <code>Component</code> that is the candidate to receive focus.
    *          
    * @param   doLeave
    *          <code>true</code> means LEAVE event will be issued
    *          
    * @param   doEntry
    *          <code>true</code> means ENTRY event will be issued
    *          
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean tryFocusChange(Widget widget, boolean doLeave,
                                         boolean doEntry)
   {
      // safety check
      if (widget == null)
         return false;
   
      Widget inFocus = UiUtils.getCurrentFocus();
      
      if (inFocus == widget)
         return false;
      
      if (inFocus != null && inFocus.isEnabled() && doLeave)
      {
         // create and process the LEAVE event first so that any trigger
         // defined for that event gets its chance to run
         KeyInput leave = new KeyInput(inFocus,
                                       EventType.KEY_PRESSED,
                                       Keyboard.SE_LEAVE);
   
         ThinClient.getInstance().processProgressEvent(leave);
   
         // check to see if a trigger returned no-apply
         if (leave.isConsumed())
            return false;
      }
   
      // if there is no ENTRY sent, focus must not be adjusted
      if (!doEntry)
         return true;
   
      // change the focus temporarily now
      widget.requestFocus();
      
      // create and process the ENTRY event now
      KeyInput entry = new KeyInput(widget,
                                    EventType.KEY_PRESSED,
                                    Keyboard.SE_ENTRY);
      ThinClient.getInstance().processProgressEvent(entry);
   
      // check to see if a trigger returned no-apply
      if (!entry.isConsumed())
         return true;
   
      // reverse the focus change
      if (inFocus != null)
         inFocus.requestFocus();
      else
         widget.parent().setFocus(null);
   
      return false;
   }

   /**
    * Performs conditional focus change procedure to the next or previous
    * widget in the tab order. The focus change can be cancelled by a
    * trigger.
    * 
    * @param   widget
    *          <code>Component</code> that is the reference widget, which is
    *          currently retaining focus.
    * @param   direction
    *          The direction of focus change. If this parameter is set to
    *          <code>true</code> then next in Tab order widget is selected.
    *          
    * @return  <code>true</code> if the focus change is successful.
    */
   public boolean focusChange(Widget widget, boolean direction)
   {
      if (widget == null)
         return false;
      
      Container<?> parent = widget.parent();
      
      // if there is no other focusable widget in the parent container, then
      // there won't be an actual focus change, just return before any events
      // are generated
      if (!parent.hasFocusable(widget))
         return false;
   
      // the list of widgets that will not receive focus;
      // this covers the case when widgets are enabled/made sensitive
      // in the LEAVE trigger; such widgets are really enabled or made
      // sensitive, they are just ignored in the tab order.
      List<Widget<?>> ignoreFocus = null;
      
      // create and process the LEAVE event first so that any trigger
      // defined for that event gets its chance to run
      
      if (widget.isEnabled())
      {
         // build the list of focusable widgets currently not traversable (not enabled or insensitive)
         List<Widget<?>> noTraverseWidgets = new ArrayList<>(parent.widgets().length);
               
         for (Widget<?> w : parent.widgets())
         {
            if (w.isFocusable() && !w.focusTraversable())
            {
               noTraverseWidgets.add(w);
            }
         }
         
         KeyInput leave = new KeyInput(widget,
                                       EventType.KEY_PRESSED,
                                       Keyboard.SE_LEAVE);
   
         tc.processProgressEvent(leave);
         
         // check to see if the trigger returned no-apply
         if (leave.isConsumed())
            return false;
         
         ignoreFocus = new ArrayList<>(noTraverseWidgets.size());
         
         // if focusable non-traversable widget was made traversable 
         // (with SENSITIVE attribute or ENABLE statement) ignore it in the tab order
         for (Widget<?> w : noTraverseWidgets)
         {
            if (w.focusTraversable())
            {
               ignoreFocus.add(w);
            }
         }

         // set the widgets that must not receive focus, see above
         parent.setIgnoreFocus(ignoreFocus);
      }
      
      try
      {
         // change the focus temporarily now
         if (direction)
         {
            parent.nextFocus();
         }
         else
         {
            parent.prevFocus();
         }
      }
      finally
      {
         // clear the widgets that must not receive focus
         parent.setIgnoreFocus(null);
      }
   
      // this is the new widget that will receive focus
      ThinClient tc = ThinClient.getInstance();
      Widget inFocus = UiUtils.getCurrentFocus();
   
      if (inFocus == widget && !widget.focusTraversable())
      {
         // if the current widget is left on focus and it is not focusable,
         // focus remains undefined. this is the case when focus LEAVE has
         // disabled current focus
         Window.getDefaultWindow().setFocus(null);
         return true;
      }
   
      // Text widgets grouping mode support - TAB/BACKTAB in this mode
      // switches groups, not individual FillIn widgets.
      if (inFocus instanceof FillIn)
      {
         FillIn fin = (FillIn) inFocus;
         
         fin = fin.getFirstInGroup();
         
         if (!fin.hasFocus())
            fin.requestFocus();
      }
      
      // restore the focus to the original widget
      setFocusOn(widget);
   
      // create and process the ENTRY event now. the ENTRY event will be 
      // responsible for setting focus to the required widget, if it accepts
      // it
      KeyInput entry = new KeyInput(inFocus,
                                    EventType.KEY_PRESSED,
                                    Keyboard.SE_ENTRY,
                                    true);
      tc.processProgressEvent(entry);
   
      // check to see if the event was consumed (due to no-apply or widget 
      // lost focusability - was disabled or hidden)
      if (!entry.isConsumed())
      {
         // pass event to widget
         inFocus.processKeyEvent(entry);
         return true;
      }
      
      // if a trigger returned no-apply, search and focus the next widget
      // in tab-order after it, without wrapping; this widget will NOT 
      // ignore no-apply
      if (tc.isTriggerResultNoApply())
      {
         Frame frame = (Frame) UiUtils.locateFrame(inFocus);
         do 
         {
            inFocus = frame.getNextEnabledWidget(inFocus, FocusManager.FOCUS_NOWRAP);
         }
         while (inFocus != null && !tc.sendEntry(inFocus));
   
         if (inFocus == null)
         {
            // in no widget found, initial widget has focus
            widget.requestFocus();
         }
      }
      // if the widget was disabled by the triggers, focus remains on 
      // initial focus
      else
      {
         // reverse the focus change
         widget.requestFocus();
      }
   
      return false;
   }

   /**
    * Get state of the {@link #noFocusAdjustment} flag.
    * 
    * @return  current value.
    */
   public boolean isNoFocusAdjustment()
   {
      return noFocusAdjustment;
   }

   /**
    * Force the initial focus to a specific widget when entering an
    * interactive/editing mode.  This will generate <code>FOCUS_GAINED</code>
    * but it will not generate a <code>LEAVE</code> or <code>FOCUS_LOST</code>
    * since it is the initial focus set.
    * <p>
    * This processing will cause a repaint so it must be bracketed by
    * invalidation processing in the caller.
    *
    * @param    widget
    *           The widget to receive focus (must not be <code>null</code>).
    */
   public static void setInitialFocus(Widget widget)
   {
      // unconditional focus change since it's the initial set and should
      // not generate a LEAVE event for the previous widget
   
      EventManager.postEvent(new FocusEvent(widget, EventType.FOCUS_GAINED));
      
      ThinClient tc = ThinClient.getInstance();
      
      if (widget.parent() != null)
         widget.parent().setFocus(widget);
      
      if (tc.getInputActive() > 0)
      {
         widget.setHighlighted(true);
      }
   
      widget.repaint();
      
      Frame frame = (Frame) UiUtils.locateFrame(widget);
      
      if (frame != null && tc.getActiveFrameInt() != frame)
         tc.setLastViewed(frame);
   }
}
